""" This module contains the representation of a show directive in ASP."""

from __future__ import annotations

from dataclasses import dataclass

from tree_sitter import Node

from mkdocstrings_handlers.asp.semantics.predicate import Predicate, ShowStatus
from mkdocstrings_handlers.asp.tree_sitter.node_kind import NodeKind


@dataclass
class Show:
    """A show directive in an ASP document."""

    predicate: Predicate
    """ The predicate supposed to be shown."""

    disable_default: bool = False
    """ If this show directive disables the default show behaviour."""

    @staticmethod
    def from_node(node: Node) -> Show | None:
        """
        Create a Show directive from the given node.

        Args:
            node: The node.

        Returns:
            The created Show directive representation if it was valid and non-empty,
            None otherwise.
        """
        match NodeKind.from_grammar_name(node.grammar_name):
            # The show_signature node has two children:
            # 1. The show directive
            # 2. The signature in form identifier/arity
            case NodeKind.SHOW_SIGNATURE:
                identifier, arity = node.children[1].text.decode("utf-8").split("/")
                return Show(Predicate(identifier, int(arity), show_status=ShowStatus.EXPLICIT), disable_default=True)

            # The show_term node has two children:
            # 1. The show directive
            # 2. The term with the function as a first child
            case NodeKind.SHOW_TERM:
                function = node.children[1].children[0]

                if function.child_count == 1:
                    # if the function has only one child, then it has no arguments
                    # this means the text is the identifier and the arity is 0
                    identifier = function.children[0].text.decode("utf-8")
                    arity = 0
                else:
                    # If the function has more than one child, then it has arguments
                    # the first child is the identifier and the second is
                    # the argument pool with terms as children
                    identifier = function.children[0].text.decode("utf-8")
                    arity = len(function.children[1].children[0].children) // 2

                return Show(Predicate(identifier, int(arity), show_status=ShowStatus.PARTIAL))

        return None



""" This module contains the LineComment class, which represents a single line comment in an ASP document."""

from __future__ import annotations

from dataclasses import dataclass

from tree_sitter import Node


@dataclass
class LineComment:
    """A line comment in an ASP document."""

    row: int
    """ The row of the line comment. """
    line: str
    """ The line of text of the comment. """

    @staticmethod
    def from_node(node: Node) -> LineComment:
        """
        Create a line comment from a node.
        """
        split_lines = node.text.decode("utf-8").split("\n")

        clean_lines = []

        for line in split_lines:
            clean_lines.append(line.removeprefix("%"))

        clean_text = "\n".join(clean_lines)

        return LineComment(row=node.start_point.row, line=clean_text)



"""
Module containing the handler for ASP files.
"""

from collections import deque
from pathlib import Path

from markupsafe import Markup
from mkdocstrings import BaseHandler, HeadingShiftingTreeprocessor

from mkdocstrings_handlers.asp.document import Document
from mkdocstrings_handlers.asp.features.dependency_graph import DependencyGraph
from mkdocstrings_handlers.asp.features.encoding_info import EncodingInfo
from mkdocstrings_handlers.asp.features.predicate_info import PredicateInfo
from mkdocstrings_handlers.asp.semantics.document_parser import DocumentParser
from mkdocstrings_handlers.asp.tree_sitter.parser import ASPParser

try:
    import tomllib  # Python 3.11+
except ModuleNotFoundError:
    import tomli as tomllib  # Use tomli for Python < 3.11

with open("pyproject.toml", "rb") as f:
    project_data = tomllib.load(f)

import logging

log = logging.getLogger(__name__)


class ASPHandler(BaseHandler):
    """MKDocStrings handler for ASP files."""

    handler = "asp"
    domain = "asp"
    name = "asp"

    # Only *extras* you want on top of mkdocstrings defaults (don’t include "autorefs" here)
    EXTRA_MDX = [
        "admonition",
        "def_list",
        "toc",
        "pymdownx.details",
        "pymdownx.superfences",
    ]

    EXTRA_MDX_CONFIG = {
        "toc": {"permalink": True},
        # If Material icons like :material-table: stop rendering, uncomment below:
        # "pymdownx.emoji": {
        #     "emoji_index": "materialx.emoji.twemoji",
        #     "emoji_generator": "materialx.emoji.to_svg",
        # },
    }

    DEFAULT_CONFIG = {
        "start_level": 1,
        "encodings": {"source": False, "git_link": False},
        "glossary": {
            "include_undocumented": True,
            "include_hidden": True,
            "include_references": True,
            "include_navigation": True,
        },
        "predicate_table": {"include_undocumented": True, "include_hidden": True},
        "dependency_graph": {"custom": True},
    }

    def __init__(self, theme: str = "material", **kwargs):
        mdx = kwargs.pop("mdx", list(self.EXTRA_MDX))
        mdx_config = kwargs.pop("mdx_config", dict(self.EXTRA_MDX_CONFIG))

        # try:
        #     import mkdocs_autorefs  # just to check it's installed

        #     mdx.append("mkdocs_autorefs")
        # except ImportError:
        #     pass
        super().__init__(theme=theme, mdx=mdx, mdx_config=mdx_config, **kwargs)

        self.env.filters["convert_markdown_simple"] = self.do_convert_markdown_simple

    def get_default_options(self) -> dict:
        """Return default options for this handler (mkdocstrings will merge user options into these)."""
        return dict(self.DEFUALT_CONFIG)

    def do_convert_markdown_simple(
        self,
        text: str,
        heading_level: int,
    ) -> Markup:
        """Render Markdown text without adding headers to the TOC

        Arguments:
            text: The text to convert.
            heading_level: The base heading level to start all Markdown headings from.

        Returns:
            An HTML string.
        """
        old_headings = [e for e in self._headings]
        treeprocessors = self._md.treeprocessors
        treeprocessors[HeadingShiftingTreeprocessor.name].shift_by = heading_level  # type: ignore[attr-defined]

        try:
            md = Markup(self._md.convert(text))
        finally:
            treeprocessors[HeadingShiftingTreeprocessor.name].shift_by = 0  # type: ignore[attr-defined]
            self._md.reset()

        self._headings = old_headings
        return md

    def parse_files(
        self, asp_parser: ASPParser, document_parser: DocumentParser, paths: list[Path]
    ) -> dict[Path, Document]:
        """
        Parse the files at the given paths and return a dictionary of documents.

        This also handles the inclusion of other files.

        Args:
            asp_parser: The ASP parser.
            document_parser: The document parser.
            paths: The paths to parse.

        Returns:
            A dictionary of documents.
        """
        parse_queue = deque(paths)
        documents: dict[Path, Document] = {}
        while parse_queue:
            path = parse_queue.popleft()
            if path.suffix != ".lp" or not path.is_file():
                continue

            with open(path, "r", encoding="utf-8") as f:
                content = f.read()

            tree = asp_parser.parse(content)
            document = document_parser.parse(Document(path, content), tree)
            documents[path] = document
            parse_queue.extend(include.path for include in document.includes if include.path not in documents)

        return documents

    def collect(self, identifier: str, config: dict) -> dict | None:
        """
        Collect data from ASP files.

        This function will be called for all markdown files annotated with '::: some/path/to/file.lp'.

        Args:
            identifier: The identifier used in the annotation.
            config: The configuration dictionary.

        Returns:
            The collected data as a dictionary.
        """
        # if identifier != "examples/my_test/base.lp":
        #     return None
        start_path = Path(identifier)
        asp_parser = ASPParser()
        document_parser = DocumentParser()

        documents: dict[Path, Document] = self.parse_files(asp_parser, document_parser, [start_path])

        data = {
            "project_name": project_data["project"]["name"],
            "project_url_tree": project_data["project"]["urls"]["Homepage"].replace(".git/", "/") + "tree/master/",
            "encodings": EncodingInfo.from_documents(documents.values()).encodings,
            "dependency_graph": DependencyGraph.from_document(documents.values()),
        }

        def get_key(x):
            key = x[1].show_status
            if x[1].show_status == 0:
                key = 3.5
            if not x[1].is_input:
                key += 0.1
            return key

        predicates = dict(
            sorted(
                PredicateInfo.from_documents(documents.values()).predicates.items(),
                key=get_key,
            )
        )
        data["predicate_info"] = predicates
        return data

    def render(self, data: dict, config: dict):
        """
        Render the collected data to HTML safely.

        Args:
            data: Collected data from the handler.
            config: Configuration dictionary.

        Returns:
            Rendered HTML string.
        """
        if not data or all(len(v) == 0 for v in data.values()):
            log.warning("\033[93mNo data to render, returning empty HTML.\033[0m")
            return "<p>No documentation available.</p>"

        # Ensure start_level is an int
        if "start_level" not in config or not isinstance(config["start_level"], int):
            config["start_level"] = self.DEFAULT_CONFIG["start_level"]

        # Merge default config safely
        sections = ["encodings", "predicate_table", "dependency_graph", "glossary"]
        for s in sections:
            if s in config:
                if isinstance(config[s], bool):
                    config[s] = {}
                default_encodings = self.DEFAULT_CONFIG[s]
                config[s] = {**default_encodings, **config[s]}

        # Initialize Jinja2 environment if not already
        if not hasattr(self, "env"):
            from jinja2 import Environment, FileSystemLoader

            self.env = Environment(loader=FileSystemLoader(os.path.join(os.path.dirname(__file__), "templates")))

        # Load template
        try:
            template = self.env.get_template("documentation.html.jinja")
        except Exception:
            return "<p>Template not found.</p>"

        # Render safely
        try:
            return template.render(**data, config=config)
        except Exception:
            return "<p>Rendering failed.</p>"

    def get_templates(self) -> Path:
        return Path(__file__).parent / "templates"



""" This module contains the Statement class, which represents a statement in an ASP program."""

from __future__ import annotations

from dataclasses import dataclass
from typing import Tuple

from tree_sitter import Node

from mkdocstrings_handlers.asp.semantics.predicate import Predicate


@dataclass
class Statement:
    """A statement in an ASP program."""

    row: int
    """The row in the source file where the statement is located."""
    text: str
    """The text of the statement."""
    provided_predicates: list[Tuple[Predicate, bool]]
    """The predicates provided by the statement."""
    needed_predicates: list[Tuple[Predicate, bool]]
    """The predicates needed by the statement."""

    @staticmethod
    def from_node(node: Node) -> Statement:
        """
        Create a statement from a node.
        Args:
            node: The node representing the statement.
        Returns:
            The created statement.
        """
        return Statement(
            row=node.start_point.row, text=node.text.decode("utf-8"), provided_predicates=[], needed_predicates=[]
        )

    def add_provided(self, predicate: Predicate, negation: bool = False):
        """Add a predicate this statement provides."""
        self.provided_predicates.append((predicate, negation))

    def add_needed(self, predicate: Predicate, negation: bool = False):
        """Add a predicate this statement needs."""
        self.needed_predicates.append((predicate, negation))



""" This module contains the NodeKind class, which represents the kind of a node in the abstract syntax tree."""

from __future__ import annotations

from enum import Enum, auto


class NodeKind(Enum):
    """The kind of a node in the abstract syntax tree."""

    UNKNOWN = auto()
    STATEMENT = "statement"
    HEAD = "head"
    BODY = "body"
    SYMBOLIC_ATOM = "symbolic_atom"
    LITERAL_TUPLE = "literal_tuple"
    LINE_COMMENT = "line_comment"
    BLOCK_COMMENT = "block_comment"
    SHOW_SIGNATURE = "show_signature"
    SHOW = "show"
    SHOW_TERM = "show_term"
    INCLUDE = "include"
    ERROR = "ERROR"

    @staticmethod
    def from_grammar_name(grammar_name: str):
        """
        Create the node kind from the given grammar name.

        This returns NodeKIind.UNKNOWN if the grammar name is not known.

        Args:
            grammar_name: The grammar name.

        Returns:
            The node kind.
        """
        return _node_kind_map.get(grammar_name, NodeKind.UNKNOWN)


_node_kind_map = {kind.value: kind for kind in NodeKind}
"""Map of node kind values to node kinds."""



"""This module contains the Predicate class, which holds information about a specific ASP predicate."""

from __future__ import annotations

import string
from dataclasses import dataclass
from enum import IntEnum

from tree_sitter import Node

from mkdocstrings_handlers.asp.semantics.predicate_documentation import PredicateDocumentation


class ShowStatus(IntEnum):
    """Enum for predicate show status with bitwise-compatible values."""

    DEFAULT = 0
    EXPLICIT = 1
    PARTIAL = 2
    PARTIAL_AND_EXPLICIT = 3
    HIDDEN = 4


@dataclass
class Predicate:
    """A predicate in an ASP document."""

    identifier: str
    """ The identifier of the predicate."""

    arity: int
    """ The arity of the predicate."""

    is_input: bool = False
    """ If it is an input (Not defined by any rule)."""

    show_status: ShowStatus = ShowStatus.DEFAULT
    """ The show status of the predicate."""

    documentation: PredicateDocumentation | None = None
    """ The documentation of the predicate."""

    @staticmethod
    def from_node(node: Node) -> Predicate:
        """
        Create a predicate from a node.

        Args:
            node: The node representing the predicate.

        Returns:
            The created predicate.
        """
        atom = node.children[0] if node.child_count == 1 else node.children[1]

        identifier = atom.children[0].text.decode("utf-8")

        if atom.child_count == 1:
            arity = 0
        else:
            terms = atom.children[1].children[0]
            arity = len(terms.children) // 2

        return Predicate(identifier, arity)

    def __str__(self) -> str:
        """
        Return the string representation of the predicate.

        If the predicate has documentation, return the representation from the documentation.
        Otherwise, return the default representation.

        The default representation is of the form `identifier(A, B, C)` where `A`, `B`, and `C` are
        the first three uppercase letters of the alphabet.

        Returns:
            The string representation of the predicate.
        """
        if self.documentation is not None:
            return self.documentation.signature
        args = ", ".join(string.ascii_uppercase[: self.arity])
        return f"{self.identifier}({args})"

    @property
    def signature(self) -> str:
        """
        Return the signature of the predicate.

        The signature of a predicate is a string of the form `identifier/arity`.

        Returns:
            The signature of the predicate.
        """
        return f"{self.identifier}/{self.arity}"

    def update_show_status(self, status: ShowStatus) -> None:
        """
        Update the show status of the predicate.

        Args:
            status: The new show status.
        """
        if self.show_status == ShowStatus.DEFAULT:
            self.show_status = status
        else:
            self.show_status |= status

        if self.show_status > ShowStatus.HIDDEN:
            self.show_status &= ShowStatus.PARTIAL_AND_EXPLICIT



""" This module contains the ASPParser class, which uses tree-sitter to parse ASP text."""

import ctypes
import os
from sys import platform

from tree_sitter import Language, Parser, Tree


class ASPParser:
    """
    A tree-sitter parser for ASP text.
    """

    def __init__(self) -> None:
        """
        Create a new ASP parser.
        """
        # Determine the path to the shared library based on the operating system
        if platform == "linux":
            lib_path = os.path.join(os.path.dirname(__file__), "lib/clingo-language.so")
        elif platform == "darwin":
            lib_path = os.path.join(os.path.dirname(__file__), "lib/clingo-language.dylib")
        else:
            lib_path = os.path.join(os.path.dirname(__file__), "lib/clingo-language.dll")

        # Load the shared library
        clingo_lib = ctypes.cdll.LoadLibrary(lib_path)

        # Retrieve the 'tree_sitter_clingo' function
        tree_sitter_clingo = clingo_lib.tree_sitter_clingo
        tree_sitter_clingo.restype = ctypes.c_void_p

        # Create a Language object using the function pointer
        clingo_language = Language(tree_sitter_clingo())

        self.parser = Parser(clingo_language)

    def parse(self, text: str) -> Tree:
        """
        Parse the given text.

        Args:
            text: The text to parse.

        Returns:
            The parsed tree.
        """
        return self.parser.parse(bytes(text, "utf8"))



"""
The main entry point for the application.
"""

import sys

from .utils.logging import configure_logging, get_logger
from .utils.parser import get_parser


def main() -> None:
    """
    Run the main function.
    """
    parser = get_parser()
    args = parser.parse_args()
    configure_logging(sys.stderr, args.log, sys.stderr.isatty())

    log = get_logger("main")
    log.info("info")
    log.warning("warning")
    log.debug("debug")
    log.error("error")


if __name__ == "__main__":
    main()



"""This module contains the classes for building a dependency graph from an ASP document."""

from __future__ import annotations

import logging
from dataclasses import dataclass

from clingo import Control

from mkdocstrings_handlers.asp.document import Document
from mkdocstrings_handlers.asp.semantics.block_comment import BlockComment
from mkdocstrings_handlers.asp.semantics.line_comment import LineComment
from mkdocstrings_handlers.asp.semantics.statement import Statement

log = logging.getLogger(__name__)


def is_clingo_code(code: str) -> bool:
    """
    Check if the given code is clingo code.

    Args:
        code: The code to check.

    Returns:
        True if the code is clingo code, False otherwise.
    """

    def silent_logger(message, _):
        pass  # Ignore all messages

    ctl = Control(["--warn=none"], logger=silent_logger)
    try:
        ctl.add("base", [], code)
        ctl.ground([("base", [])])

        return True
    except Exception:
        return False


class EncodingLineType:
    """Type of the encoding line."""

    CODE = "code"
    """Code line"""
    MD = "md"
    """Markdown line"""


@dataclass
class EncodingLine:
    """Line in the encoding."""

    type: str
    """Wither code or md"""
    str_content: str
    """Content of the line"""


@dataclass
class Encoding:
    source: str
    """Raw source code of the encoding."""
    content: list[EncodingLine]
    statements: list[Statement]

    predicates: list[str]


@dataclass
class EncodingInfo:
    """Content of the encoding including statements and lines."""

    encodings: dict[str, Encoding]

    @staticmethod
    def from_documents(documents: list[Document]) -> EncodingInfo:
        """
        Create a encoding content from an ASP document.

        Args:
            document: The ASP document.

        Returns:
            The encoding content.
        """
        encodings: dict[str, Encoding] = {}

        for document in documents:
            lines: list[EncodingLine] = []

            for oo in document.ordered_objects:
                if isinstance(oo, Statement):
                    if lines and lines[-1].type == EncodingLineType.CODE:
                        lines[-1].str_content += "\n" + oo.text
                    else:
                        lines.append(EncodingLine(EncodingLineType.CODE, oo.text))
                if isinstance(oo, BlockComment):
                    content = "\n".join(oo.lines)
                    lines.append(EncodingLine(EncodingLineType.MD, content))
                if isinstance(oo, LineComment):
                    is_code = is_clingo_code(oo.line)
                    if not is_code:
                        lines.append(EncodingLine(EncodingLineType.MD, oo.line))
                    else:
                        log.debug("Commented code ignored:", oo.line)

            encodings[document.path] = Encoding(
                source=document.content,
                content=lines,
                statements=document.statements,
                predicates=[predicate.signature for predicate in document.predicates.values()],
            )

        return EncodingInfo(encodings)



"""This module contains the 'PredicateDocumentation', which represents the documentation for a predicate."""

from __future__ import annotations

import re
import textwrap
from dataclasses import dataclass, field

from tree_sitter import Node

from mkdocstrings_handlers.asp.semantics.block_comment import BlockComment
from mkdocstrings_handlers.asp.tree_sitter.node_kind import NodeKind
from mkdocstrings_handlers.asp.tree_sitter.parser import ASPParser
from mkdocstrings_handlers.asp.tree_sitter.traverse import traverse


@dataclass
class PredicateParameterDocumentation:
    """
    Documentation for a parameter of a predicate.
    """

    name: str
    """ The name of the parameter. """
    type: str | None = None
    """ The type of the parameter. """
    description: str = ""
    """ The description of the parameter. """

    def __str__(self) -> str:
        return f"{self.name} ({self.type}) : {self.description}"


@dataclass
class PredicateDocumentation:
    """
    Documentation for a predicate.


    Example:
        %*#some_predicate(A,B,C).
        description
        #parameters
        - A : this is  A
        - B : this is  B
        - C : this is  C
        *%
    """

    signature: str
    """ The signature of the predicate. """
    description: str
    """ The description of the predicate. """
    parameter_documentations: dict[str, PredicateParameterDocumentation] = field(default_factory=dict)
    """ The documentation of the parameters of the predicate. """
    node: Node | None = None
    """ The node representing the predicate. """

    @staticmethod
    def from_block_comment(comment: BlockComment) -> PredicateDocumentation | None:
        """
        Create a predicate documentation from a comment.

        Args:
            comment: The block comment.

        Returns:
            The predicate documentation or None if the comment is not a predicate documentation.
        """

        signature_re = r"(?P<signature>.*?)\n\s*\.{5,}\n"
        description_re = r"(?P<description>.*?)Args:"
        args_re = r"(?P<args>.*?(?=\n\S|\Z))"

        docstring_re = re.compile(
            signature_re + description_re + args_re,
            re.DOTALL | re.MULTILINE,
        )

        match = docstring_re.match(comment.text)

        if not match:
            return None

        # Get the signature
        signature = match.group("signature").strip()

        # Parse the signature to get the literal
        predicate_node = None

        def identifier_from_node(node: Node):
            nonlocal predicate_node
            if NodeKind.from_grammar_name(node.grammar_name) == NodeKind.SYMBOLIC_ATOM:
                predicate_node = node.parent

        parser = ASPParser()
        tree = parser.parse(f"{signature}.")
        traverse(tree, identifier_from_node, lambda _: None)

        description = textwrap.dedent(match.group("description")).strip()
        args = textwrap.dedent(match.group("args"))

        args_name = r"(?P<name>\w+)"
        args_type = r"(?:\s*\((?P<type>[^)]+)\))?"
        args_description = r"\s*:\s*(?P<description>.+?)"

        args_re = rf"^{args_name}{args_type}\s*{args_description}(?=(?:^\S|\Z))"

        args_matches = re.finditer(args_re, args, re.MULTILINE | re.DOTALL)

        parameter_documentations: dict[str, PredicateParameterDocumentation] = {}

        for args_match in args_matches:
            arg_name = args_match.group("name").strip()
            arg_type = args_match.group("type").strip() if args_match.group("type") else ""
            arg_description = args_match.group("description").strip()

            parameter_documentations[arg_name] = PredicateParameterDocumentation(
                name=arg_name,
                type=arg_type,
                description=arg_description,
            )

        return PredicateDocumentation(
            signature=signature,
            description=description,
            parameter_documentations=parameter_documentations,
            node=predicate_node,
        )



"""Entry point of the mkdocstrings handler module."""

from .handler import ASPHandler


def get_handler(**kwargs):
    """
    Return an instance of the ASPHandler class.

    This is required by mkdocstrings to load the handler.
    """

    return ASPHandler(**kwargs)



"""Utility functions for debugging tree-sitter trees."""

from tree_sitter import Node


def print_tree(node: Node, source: bytes, depth: int) -> None:
    """Recursively print the tree structure of a node.
    Args:
        node: The node to print.
        source: The source code from which the node was created.
        depth: The current depth in the tree.
    """
    node_text = node.text.decode("utf-8")
    print(f"{'  ' * depth}{node.grammar_name} {node.id}: {node_text}")

    for i in range(node.child_count):
        child = node.child(i)
        print_tree(child, source, depth + 1)
